#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SDL.h"
#include "SDL_image.h"
#include "SDL_endian.h"

#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"

#include "fcgame.h"

/////map format is big endian, because i like it better.
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
#define SWAP16(X)    (X)
#define SWAP32(X)    (X)
#else
#define SWAP16(X)    SDL_Swap16(X)
#define SWAP32(X)    SDL_Swap32(X)
#endif


#define engine_version(x, y, z) ((x << 16) + (y << 8) + z)
#define MAP_MAGIC 0xFC00

typedef struct{//Map file header
	unsigned short mapmagic;   //magic number (FC00)
	unsigned long version;     //engine version.
	unsigned short header_length;
	char name[16];   //Map name.
	char tileset[16];   //filename of tileset used for map
	unsigned short mapwidth;
	unsigned short mapheight;
	char scripts[16];   //filename for lua scripts
} map_header;

unsigned short get_tileid(unsigned short *layer, signed short x, signed short y){
	if(x >= mapwidth || y >= mapheight || x < 0 || y < 0){
		return 0;
	}
	return layer[((y*mapwidth)+x)];
}


entity_t *get_entity(signed short x, signed short y){
	if(x >= mapwidth || y >= mapheight || x < 0 || y < 0){
		return NULL;
	}
	entity_t *entity = layer_entities[((y*mapwidth)+x)];
	return entity;
}


void entity_face(entity_t *entity, int direction){
	entity->direction = direction;
	
	char *luatring = malloc(160);
	sprintf(luatring, "%s:setdirection(%u)", entity->name, entity->direction);
	lua_dostring(luaVM, luatring);
}

typedef struct{
	char name[16];
	unsigned short tileid;//START tile
	unsigned short x,y;
	unsigned char direction;
	unsigned char flags; //useless ATM
} mapfile_entity __attribute__((packed));

void unloadmap(){
	free_entities();
	
	if(layer_bottom != NULL)
		free(layer_bottom);
		
	if(layer_interact != NULL)
		free(layer_interact);
	
	if(layer_top != NULL)
		free(layer_top);
	
	if(layer_entities != NULL)
		free(layer_entities);
	
	if(tileset != NULL){
		SDL_FreeSurface(tileset);
		tileset == NULL;
	}
	
	if(player != NULL){
		free(player);
		player = NULL;
	}
}

bool loadmap(char *filename){
	unloadmap();
	
	strcpy(current_map, (const char *)filename);
	
	FILE *map;
	
	map = fopen(filename, "rb");
	if(map == NULL){
		printf("Unable to open %s", filename);
 		exit(0);
	}
	
	map_header header;
	
	fread(&header, sizeof(map_header), 1, map);
	
	#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		header.mapmagic = SDL_Swap16(header.mapmagic);
		header.header_length = SDL_Swap16(header.header_length);
		header.version = SDL_Swap32(header.version);
		header.mapwidth = SDL_Swap16(header.mapwidth);
		header.mapheight = SDL_Swap16(header.mapheight);
	#endif
	
	if(header.mapmagic != MAP_MAGIC){
		printf("Bad magic number in map file (0x%x)\n", header.mapmagic);
		return false;
	}
	
	if(header.header_length < sizeof(map_header)){
		printf("WARNING: map's header length (%u) is under %u\n", header.header_length, sizeof(map_header));
	}
	printf("========================================\n");
	
	printf("%s\n", header.name);
	printf("version: %u.%u.%u\n", (int)(header.version << 16)&0xff, (int)(header.version << 8)&0xff, (int)(header.version)&0xff);
	printf("tileset: %s\n", header.tileset);
	printf("%ux%u\n", header.mapwidth, header.mapheight);
	
	tileset = image_load(header.tileset);
	
	if (tileset == NULL) {
		fprintf(stderr, "ERROR: cannot load tileset \"%s\"\n", header.tileset);
		exit(1);
	}
	
	mapwidth = header.mapwidth;
	mapheight = header.mapheight;
	
	layer_bottom = malloc(mapwidth * mapheight * sizeof(unsigned short));
	layer_top = malloc(mapwidth * mapheight * sizeof(unsigned short));
	layer_interact = malloc(mapwidth * mapheight * sizeof(unsigned short));
	layer_entities = malloc(mapwidth * mapheight * sizeof(void *));
	
	//////////////////////////////////////////////////////////////////////////////////
	
	fseek(map, header.header_length+1, 0);
	
	fread(layer_bottom, mapwidth * mapheight * sizeof(unsigned short), 1, map);
	
	if(feof(map)){
		printf("ERROR: Unexpected End Of File!\n");
		
		return false;
	}
	
	fread(layer_interact, mapwidth * mapheight * sizeof(unsigned short), 1, map);
	
	if(feof(map)){
		printf("ERROR: Unexpected End Of File!\n");
		
		return false;
	}
	
	fread(layer_top, mapwidth * mapheight * sizeof(unsigned short), 1, map);
	
	if(feof(map)){
		printf("ERROR: Unexpected End Of File!\n");
		
		return false;
	}
	
	//////////////////////////////////////////////////////////////////////////////////
	
	for(unsigned long i=0;i<(mapwidth*mapheight);i++){
		#if SDL_BYTEORDER == SDL_LIL_ENDIAN
			layer_bottom[i] = SDL_Swap16(layer_bottom[i]);
			layer_top[i] = SDL_Swap16(layer_top[i]);
			layer_interact[i] = SDL_Swap16(layer_interact[i]);
		#endif
		
		layer_entities[i] = (entity_t *)NULL;
	}
	
	player = NULL;
	
	printf("%s\n", header.scripts);
	scripting_loadfile(header.scripts);
	
	fclose(map);
	
	if(player == NULL){
		fprintf(stderr, "ERROR: %s: end reached without call of setup_player\n", header.scripts);
		exit(0);
	}
	
	reload_entities();
	
	return true;
}

void free_entities(){
	if(layer_entities == NULL){
		return;
	}
	for(unsigned long i=0;i<(mapwidth*mapheight);i++){
		if(layer_entities[i] != NULL){
			if(layer_entities[i] != player){
				free(layer_entities[i]);
			}
			layer_entities[i] = NULL; //to be safe
		}
	}
}

void reload_entities(){
	lua_dostring(luaVM, "reload_entities()");
}

int setup_player(lua_State* luaS){
	if(player == NULL){
		player = (entity_t *)malloc(sizeof(entity_t));
		whatplayer = 5;
	}else{
		layer_entities[(((player->y)*mapwidth)+(player->x))] = NULL;
	}
	
	player->tileid = players_tile_start;
	player->x = lua_tonumber(luaVM, 1);
	player->y = lua_tonumber(luaVM, 2);
	player->realx = (player->x)*TILEWIDTH;
	player->realy = (player->y)*TILEHEIGHT;
	
	player->direction = lua_tonumber(luaVM, 3);
	
	player->step = 0;

	
	layer_entities[(((player->y)*mapwidth)+(player->x))] = player;
	
	lua_pushnumber(luaS, 0);
	return 1;
}

int create_entity(lua_State* luaS){
	char *name = (char *)lua_tostring(luaS, 1);
	//char *name = lua_tostring(luaVM, 1);
	int x = lua_tonumber(luaVM, 2);
	int y = lua_tonumber(luaVM, 3);
	int tileid = lua_tonumber(luaVM, 4);
	int direction = lua_tonumber(luaVM, 5);
	int flags = lua_tonumber(luaVM, 6);
	
	/*printf("=========================\n");
	printf("name: %s\n", name);
	printf("x: %u\n", x);
	printf("y: %u\n", y);
	printf("tileid: %u\n", tileid);
	printf("direction: %u\n", direction);
	printf("flags: %u\n", flags);*/
	
	if(get_entity(x,y) != NULL){
		fprintf(stderr, "ERROR: (%u,%u): cannot place entities on top of one another\n", x, y);
		lua_pushnumber(luaS, -1);
		return 0;
	}
	
	entity_t *entity = (entity_t *)malloc(sizeof(entity_t));
	strncpy(entity->name, name, 16);
	
	entity->tileid = tileid;
	entity->x = x;
	entity->y = y;
	entity->direction = direction;
	
	entity->step = 0;
	entity->realx = (entity->x)*TILEWIDTH;
	entity->realy = (entity->y)*TILEHEIGHT;
	
	layer_entities[(((entity->y)*mapwidth)+(entity->x))] = entity;
	
	lua_pushnumber(luaS, 0);
	return 1;
}






